#include "renderer.h"
#include <stdio.h>

// STB_IMAGE for loading images of many filetypes
#ifdef _MSC_VER
#pragma warning(disable:4996)
#endif
#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>
#define STB_IMAGE_WRITE_IMPLEMENTATION
#include <stb_image_write.h>

SDL_Window*   gWindow    = NULL; //The window we'll be rendering to
SDL_Renderer* gRenderer  = NULL; //The window renderer
TTF_Font*     gFont      = NULL; //Used font. Documentation of TTF_Font can be found here:  
                                 //  https://www.libsdl.org/projects/docs/SDL_ttf/SDL_ttf.html#SEC6

//Render just a part of the GfxObject's image to screen - defined in srcRect
void renderGfxSubFrame(GfxObject* gfx, int x, int y, double angle, float scale, SDL_Rect srcRect )
{
	//SDL_Rect srcRect = { 0, 0, gfx->textureWidth, gfx->textureHeight }; // select the whole image
    
    // Use scale to compute new output width, height and x,y. 
    // Note on scaling: integer-discretized output positions means that width and height should be modified 
    // an equal amount on the left/right/top/bottom sides to look good (i.e., in steps of two).
    int w = (int)(gfx->outputWidth*scale + 0.5); // round to nearest
    if ( ((w - gfx->outputWidth) % 2) == 1)
        w--;
    int h = (int)(gfx->outputHeight*scale + 0.5); // round to nearest
    if ( ((h - gfx->outputHeight) % 2) == 1)
        h--;        

    float dstX = (float)(x - w/2);
    float dstY = (float)(y - h/2);
	SDL_Rect dstRect = { (int)dstX, (int)dstY, (int)w,  (int)h };  // output position and width/height
    
	//Render to screen
    SDL_RenderCopyEx(gRenderer, gfx->mTexture, &srcRect, &dstRect, angle, NULL, SDL_FLIP_NONE);
    // NULL here specifies that rotation will be done around the center of the image. 
    // See here for details: https://wiki.libsdl.org/SDL_RenderCopyEx
}

//Render gfx object to screen
void renderGfxObject(GfxObject* gfx, int x, int y, double angle, float scale )
{
	SDL_Rect srcRect = { 0, 0, gfx->textureWidth, gfx->textureHeight }; // select the whole image
    
    // Use scale to compute new output width, height and x,y. 
    // Note on scaling: integer-discretized output positions means that width and height should be modified 
    // an equal amount on the left/right/top/bottom sides to look good (i.e., in steps of two).
    int w = (int)(gfx->outputWidth*scale + 0.5); // round to nearest
    if ( ((w - gfx->outputWidth) % 2) == 1)
        w--;
    int h = (int)(gfx->outputHeight*scale + 0.5); // round to nearest
    if ( ((h - gfx->outputHeight) % 2) == 1)
        h--;        

    float dstX = (float) (x - w/2);
    float dstY = (float) (y - h/2);
	SDL_Rect dstRect = { (int)dstX, (int)dstY, (int)w,  (int)h };  // output position and width/height
    
	//Render to screen
    SDL_RenderCopyEx(gRenderer, gfx->mTexture, &srcRect, &dstRect, angle, NULL, SDL_FLIP_NONE);
    // NULL here specifies that rotation will be done around the center of the image. 
    // See here for details: https://wiki.libsdl.org/SDL_RenderCopyEx
}

GfxObject createGfxObject(  const char* filename )
{
    GfxObject gfx;
	// init to something
	gfx.textureWidth = gfx.textureHeight = gfx.outputWidth = gfx.outputHeight = 0;
	gfx.mTexture = 0;

    //Load image at specified path
    unsigned char* data = NULL;
    SDL_Surface* loadedSurface = IMG_Load( filename );
    if( loadedSurface == NULL ) { // Try loading with stb instead of SDL (e.g., JPG for Emscripten)
        // req_format = the color format you request stb_image to output,
        // use STBI_rgb if you don't want/need the alpha channel
        int req_format = STBI_rgb_alpha;
        int width, height, orig_format;
        //stbi_set_flip_vertically_on_load(true);
        data = stbi_load(filename, &width, &height, &orig_format, req_format);
        if(data == NULL) {
          printf( "Unable to load image %s! SDL_image Error: %s\n", filename, IMG_GetError() );
          SDL_Log("Loading image failed: %s", stbi_failure_reason());
          return gfx;
        }

        // Set up the pixel format color masks for RGB(A) byte arrays.
        // Only STBI_rgb (3) and STBI_rgb_alpha (4) are supported here!
        Uint32 rmask, gmask, bmask, amask;
        #if SDL_BYTEORDER == SDL_BIG_ENDIAN
          int shift = (req_format == STBI_rgb) ? 8 : 0;
          rmask = 0xff000000 >> shift;
          gmask = 0x00ff0000 >> shift;
          bmask = 0x0000ff00 >> shift;
          amask = 0x000000ff >> shift;
        #else // little endian, like x86
          rmask = 0x000000ff;
          gmask = 0x0000ff00;
          bmask = 0x00ff0000;
          amask = (req_format == STBI_rgb) ? 0 : 0xff000000;
        #endif

        int depth, pitch;
        if (req_format == STBI_rgb) {
          depth = 24;
          pitch = 3*width; // 3 bytes per pixel * pixels per row
        } else { // STBI_rgb_alpha (RGBA)
          depth = 32;
          pitch = 4*width;
        }
        loadedSurface = SDL_CreateRGBSurfaceFrom((void*)data, width, height, depth, pitch,
                                                     rmask, gmask, bmask, amask);
        if (loadedSurface == NULL) {
          SDL_Log("Creating surface failed: %s", SDL_GetError());
          stbi_image_free(data);
          return gfx;
        }  
    }

    //Create texture from surface pixels
    gfx.mTexture = SDL_CreateTextureFromSurface( gRenderer, loadedSurface );
    if( gfx.mTexture == NULL )
        printf( "Unable to create texture from %s! SDL Error: %s\n", filename, SDL_GetError() );
    else {
        //Get image dimensions
        gfx.textureWidth  = loadedSurface->w;
        gfx.textureHeight = loadedSurface->h;
        // Set some default output size. For now, use the original image size.
        gfx.outputWidth  = gfx.textureWidth;
        gfx.outputHeight = gfx.textureHeight;
        
        SDL_SetTextureBlendMode(gfx.mTexture, SDL_BLENDMODE_BLEND);
    }
    SDL_FreeSurface( loadedSurface ); //Get rid of old loaded surface
    // .. *and* the data used by the surface!
    if(data) stbi_image_free(data);  
    
    return gfx;
}

void freeGfxObject(GfxObject* gfx)
{
	//Free texture if it exists
	if( gfx->mTexture != NULL )
		SDL_DestroyTexture( gfx->mTexture );
    gfx->mTexture = NULL; // good way to mark that the pointer is now invalid
}

bool initRenderer(int windowWidth, int windowHeight)
{
	//Initialization flag
	bool success = true;

	//Initialize SDL
	if( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK | SDL_INIT_GAMECONTROLLER | SDL_INIT_TIMER ) < 0 ) // SDL_INIT_EVERYTHING
	{
		printf( "SDL could not initialize! SDL Error: %s\n", SDL_GetError() );
		success = false;
	}
	else
	{
		//Set texture filtering to linear
		if( !SDL_SetHint( SDL_HINT_RENDER_SCALE_QUALITY, "1" ) )
		{
			printf( "Warning: Linear texture filtering not enabled!" );
		}

		//Create window
		gWindow = SDL_CreateWindow( "SDL Project", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, windowWidth, windowHeight, SDL_WINDOW_SHOWN );
		if( gWindow == NULL )
		{
			printf( "Window could not be created! SDL Error: %s\n", SDL_GetError() );
			success = false;
		}
		else
		{
			//Create vsynced renderer for window
			gRenderer = SDL_CreateRenderer( gWindow, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC );
			if( gRenderer == NULL )
			{
				printf( "Renderer could not be created! SDL Error: %s\n", SDL_GetError() );
				success = false;
			}
			else
			{
				//Initialize renderer color
				SDL_SetRenderDrawColor( gRenderer, 0x00, 0x00, 0x00, 0xFF );

				//Initialize PNG loading
				int imgFlags = IMG_INIT_PNG | IMG_INIT_JPG | IMG_INIT_TIF;
				if( !( IMG_Init( imgFlags ) & imgFlags ) )
				{
					printf( "SDL_image could not initialize! SDL_image Error: %s\n", IMG_GetError() );
					success = false;
				}
                //Initialize SDL_ttf
                if( TTF_Init() == -1 )
                {
                    printf( "SDL_ttf could not initialize! SDL_ttf Error: %s\n", TTF_GetError() );
                    success = false;
                }
                //Initialize SDL_mixer
                int frequency = 44100/4;
#if EMSCRIPTEN
                frequency = EM_ASM_INT_V({
                    var context;
                    try {
                        context = new AudioContext();
                    } catch (e) {
                        context = new webkitAudioContext();
                    }
                    return context.sampleRate;
                });          
#endif
                printf("Sample rate: %d\n", frequency);
                Mix_AllocateChannels(16);
                if( Mix_OpenAudio( frequency, MIX_DEFAULT_FORMAT, 2, 512 ) == -1 )
                {
                    printf( "SDL_mixer (i.e,. audio) could not initialize! SDL_mixer Error: %s\n", Mix_GetError() );
                    success = false;    
                }
			}
		}
	}

	return success;
}

void closeRenderer()
{
    	
    // free the font
    TTF_CloseFont(gFont);
    gFont = NULL; // For our own sake to mark that the font is closed.

    //Destroy window	
    SDL_DestroyRenderer( gRenderer );
    SDL_DestroyWindow( gWindow );
    gWindow = NULL;
    gRenderer = NULL;

    //Quit SDL subsystems
    TTF_Quit();
    IMG_Quit();
    SDL_Quit();
}


// Render Text
void renderText(const char* str, int x, int y)
{
    SDL_Color color={0xff, 0xff, 0x00}; // yellow 
    renderTextWithColor(color, str, x, y);
}
void renderTextWithColor(SDL_Color color, const char* str, int x, int y)
{
    static bool bFirstTime = true;
    if(bFirstTime)
    {
        bFirstTime = false;
        gFont = TTF_OpenFont( "assets/freefonts/FreeSans.ttf", 28 );
        if( gFont == NULL )
        {
            printf( "Failed to load font! SDL_ttf Error: %s\n", TTF_GetError() );
        }        
    }
    
    SDL_Surface *surface;
    if(!(surface=TTF_RenderText_Blended(gFont, str, color))) {
        printf("TTF_RenderText_Solid Error: %s\n", TTF_GetError());
    } 
    else 
    {
        SDL_Texture* texture = SDL_CreateTextureFromSurface(gRenderer, surface);
        SDL_Rect dstRect = {x, y, surface->w, surface->h}; // x,y,w,h
        SDL_RenderCopy(gRenderer, texture, NULL, &dstRect);
        
        SDL_FreeSurface(surface);
        SDL_DestroyTexture(texture);
    }  
}

// Just a helper function.
bool playMusic(Mix_Music *music, const char* file, int volume)
{
    if( !music ) {
        music = Mix_LoadMUS( file ); // load new song data
        if( music == NULL ) {
            printf( "Failed to load song %s! SDL_mixer Error: %s\n", file, Mix_GetError() );
            return false; // skip further attempts to load this song
        }
    } 
    Mix_VolumeMusic(volume);
    if ( Mix_PlayMusic(music, -1) == -1 ) // param "-1" -> loop forever
        printf("Mix_PlayMusic: %s\n", Mix_GetError());
    // Mix_FreeMusic(music); // at some point, we should cleanup (at end of program)
    return true;
}

// Just a helper function.  
bool playSoundEffect(Mix_Chunk * wave, const char* file, int volume, int channel)
{
    if( !wave ) {
        wave = Mix_LoadWAV( file ); // load wave
        if ( wave == NULL ) {
            printf( "Failed to load wave %s! SDL_mixer Error: %s\n", file, Mix_GetError() );
            return false;
        }
    }
    Mix_VolumeChunk(wave, volume);
    Mix_PlayChannel( channel, wave, 0 ); // params: "any free channel", chunk, "zero repetitions"
    // Mix_FreeChunk(wave); // at some point, we should cleanup (at end of program)
    return true;
}

void playSong1()
{
    static Mix_Music *music = NULL; // The background music that will be played
    static const char* file = "assets/Airwolf.ogg"; // .mp3 not supported in Emscripten
    static bool ok = true;
    int volume = MIX_MAX_VOLUME/2; // max = 128, min = 0.
    if (ok) ok = ok && playMusic(music, file, volume); // trick to let game continue without sound gracefully on fail.
}

void playShot()
{
    static Mix_Chunk *sound = NULL; // will hold the loaded sound data
    static const char* file = "assets/shot.wav";
    static bool ok = true;
    // MIX_MAX_VOLUME = 128. Min = 0.
    // channel = -1 -> first free channel. But better use a fixed channel (below: "1") per sound, to avoid dropping fast-repeting sounds.
    if (ok) ok = ok && playSoundEffect(sound, file, MIX_MAX_VOLUME/2, 1); // trick to let game continue without sound gracefully on fail.
}

void playExplosion()
{
    static Mix_Chunk *sound = NULL; // will hold the loaded sound data
    static const char* file = "assets/explosion.wav";
    static bool ok = true;
    if (ok) ok = ok && playSoundEffect(sound, file, MIX_MAX_VOLUME/2, 2); 
}

